; 
    include \masm32\include\masm32rt.inc
; 
    .data
      gdwTlsIndex dd 0
      hThread1    dd 0
      hThread2    dd 0
      hThread3    dd 0
      hThread4    dd 0
      cksum1      dd 0
      cksum2      dd 0
      cksum3      dd 0
      cksum4      dd 0
    .code
; 

DEFAULT_SEED = 12345678

align 4

;--------------------------------------------------
; This is the MASM32 nrandom procedure modified to
; use TLS for the seed.
;--------------------------------------------------

nrandom_tls PROC base:DWORD

    ;--------------------------------------------------
    ; If no thread in the process has already done so,
    ; allocate a TLS index and save the return value.
    ;--------------------------------------------------

    or gdwTlsIndex, 0
    jnz @F
    call TlsAlloc
    mov gdwTlsIndex, eax

  @@:

    ;-------------------------------------------------------------
    ; For this application the seed is stored directly in the TLS
    ; slot for the thread. If the TLS slot is in the initialized
    ; state (NULL), use the default seed. Note that nrandom will
    ; not work with a starting seed of zero (unless you intend to
    ; generate a random sequence of zeros :)).
    ;-------------------------------------------------------------

    invoke TlsGetValue, gdwTlsIndex
    test eax, eax
    jnz @F
    mov eax, DEFAULT_SEED
  @@:

  ; ****************************************
    test eax, 80000000h
    jz  @F
    add eax, 7fffffffh
  @@:
  ; ****************************************

    xor edx, edx
    mov ecx, 127773
    div ecx
    mov ecx, eax
    mov eax, 16807
    mul edx
    mov edx, ecx
    mov ecx, eax
    mov eax, 2836
    mul edx
    sub ecx, eax
    xor edx, edx
    mov eax, ecx
    div base

    mov eax, edx

    ;------------------------------------------
    ; Save the new seed, preserving the return
    ; value around the TlsSetValue call.
    ;------------------------------------------

    push eax
    invoke TlsSetValue, gdwTlsIndex, ecx
    pop eax

    ret

nrandom_tls ENDP

; 

;------------------------------------------------
; This is the MASM32 nseed procedure modified to
; use TLS for the seed.
;
; Note that since this function sets a per-thread
; seed, to be effective it must be called from
; the thread that will use the seed.
;------------------------------------------------

nseed_tls proc TheSeed:DWORD

    ;--------------------------------------------------
    ; If no thread in the process has already done so,
    ; allocate a TLS index and save the return value.
    ;--------------------------------------------------

    or gdwTlsIndex, 0
    jnz @F
    call TlsAlloc
    mov gdwTlsIndex, eax

  @@:

    ;-------------------------
    ; Set the new seed value.
    ;-------------------------

    invoke TlsSetValue, gdwTlsIndex, TheSeed

    ret

nseed_tls endp

; 

;----------------------------------------------------------
; Each of the thread procedures calculates a simple 32-bit
; checksum, as a convenient and reasonably reliable method
; of comparing the sequences.
;----------------------------------------------------------

N = 10000000

ThreadProc1 proc uses ebx junk:DWORD
    xor ebx, ebx
    .WHILE ebx < N
      invoke nrandom, 1000
      add cksum1, eax
      inc ebx
    .ENDW
    ret
ThreadProc1 endp

; 

ThreadProc2 proc uses ebx junk:DWORD
    xor ebx, ebx
    .WHILE ebx < N
      invoke nrandom, 1000
      add cksum2, eax
      inc ebx
    .ENDW
    ret
ThreadProc2 endp

; 

ThreadProc3 proc uses ebx junk:DWORD
    xor ebx, ebx
    .WHILE ebx < N
      invoke nrandom_tls, 1000
      add cksum3, eax
      inc ebx
    .ENDW
    ret
ThreadProc3 endp

; 

ThreadProc4 proc uses ebx junk:DWORD
    xor ebx, ebx
    .WHILE ebx < N
      invoke nrandom_tls, 1000
      add cksum4, eax
      inc ebx
    .ENDW
    ret
ThreadProc4 endp

; 
start:
; 

    ;-------------------------------------------
    ; Calc and display the checksum for nrandom
    ; starting with the default seed.
    ;-------------------------------------------

    invoke ThreadProc1, 0
    print ustr$(cksum1),13,10,13,10

    ;----------------------------------------
    ; Restore the global nrandom seed to its
    ; default value and cksum1 to zero.
    ;----------------------------------------

    invoke nseed, DEFAULT_SEED
    mov cksum1, 0

    ;----------------------------------------------------------------
    ; Create the threads and save the handles. Note that the threads
    ; will run immediately, and because of the high loop counts each
    ; thread will span multiple timeslices, so the execution of the
    ; threads will effectively overlap.
    ;----------------------------------------------------------------

    invoke CreateThread, 0, 0, ThreadProc1, 0, 0, 0
    mov hThread1, eax
    invoke CreateThread, 0, 0, ThreadProc2, 0, 0, 0
    mov hThread2, eax
    invoke CreateThread, 0, 0, ThreadProc3, 0, 0, 0
    mov hThread3, eax
    invoke CreateThread, 0, 0, ThreadProc4, 0, 0, 0
    mov hThread4, eax

    ;--------------------------------------------
    ; Wait until all of the threads have exited.
    ;--------------------------------------------

    invoke WaitForMultipleObjects, 4, ADDR hThread1, TRUE, INFINITE

    ;------------------------
    ; Display the checksums.
    ;------------------------

    print ustr$(cksum1),13,10
    print ustr$(cksum2),13,10,13,10
    print ustr$(cksum3),13,10
    print ustr$(cksum4),13,10,13,10

    ;-----------------------
    ; Release the TLS index.
    ;-----------------------

    invoke TlsFree, gdwTlsIndex

    inkey "Press any key to exit..."
    exit
; 
end start
